import java.io.*;
import java.net.URL;
import java.nio.file.*;
import java.util.*;

public class Preprocess {
    public static void main(String[] args) throws Exception{
	if(args.length!=2){
	    System.err.println("usage: java -cp STAR3D.jar Preprocess PDB chain");
	    System.exit(0);
	}
	
	String PDBID=args[0];
	String chainID=args[1];
		
	//get the path STAR3D.jar
	String STAR3D_PATH = new File(Preprocess.class.getProtectionDomain().getCodeSource().getLocation().getPath()).getParentFile().getPath();
	
	File PDB_DATA_PATH=new File(STAR3D_PATH, "PDB"); 
	File SI_DATA_PATH=new File(STAR3D_PATH, "STAR3D_struct_info");
	
	if(!(PDB_DATA_PATH.exists() && PDB_DATA_PATH.isDirectory())) PDB_DATA_PATH.mkdir();
	if(!(SI_DATA_PATH.exists() && SI_DATA_PATH.isDirectory())) SI_DATA_PATH.mkdir();
		
	PDBID=PDBID.toLowerCase();
	//chain ID can be lower case
	//chainID=chainID.toUpperCase();
		
	File PDB_fn=new File(PDB_DATA_PATH, PDBID+".pdb");
	
	if(!PDB_fn.exists() || ! PDB_fn.isFile()){
	    System.out.println("Downloading PDB file "+PDBID+".pdb...");
	    URL link=new URL("http://www.rcsb.org/pdb/files/"+PDBID+".pdb");
	    InputStream in =new BufferedInputStream(link.openStream());
	    ByteArrayOutputStream out=new ByteArrayOutputStream();
	    byte[] buf=new byte[1024];
	    int n=0;
	    while(-1!=(n=in.read(buf))) out.write(buf, 0, n);
	    out.close();
	    in.close();
	    byte[] response=out.toByteArray();
	    
	    FileOutputStream fos=new FileOutputStream(PDB_fn);
	    fos.write(response);
	    fos.close();
	}
		
	File MCA_fn=new File(SI_DATA_PATH, PDBID+".mca");
	if(!MCA_fn.exists() || MCA_fn.length()==0){
	    System.out.println("Processing "+PDBID+".pdb with MCAnnotate...");
	    Process p=Runtime.getRuntime().exec(new File(STAR3D_PATH, "tools/MC-Annotate").toString()+" "+PDB_fn);
	    BufferedReader in=new BufferedReader(new InputStreamReader(p.getInputStream()));
	    PrintWriter out=new PrintWriter(MCA_fn);
	    String line;
	    while((line=in.readLine())!=null) out.println(line);
	    out.close();
	    in.close();
	    p.waitFor();
	}
		
	File npk_ct_fn=new File(SI_DATA_PATH, PDBID+'_'+chainID+".npk.ct");
	if(!npk_ct_fn.exists() || npk_ct_fn.length()==0){
	    System.out.println("Removing pseudoknots from "+PDBID+"_"+chainID+"...");
	    MCA mca_parser=new MCA(MCA_fn);
	    PDB pdb_parser=new PDB(PDB_fn);
	    String chain_seq=pdb_parser.get_chain_seq(chainID.charAt(0));
	    //get the base pairs
	    HashSet<Pair<Integer, String>> chain_bp=Lib.get_seq_bp(pdb_parser, mca_parser, chainID.charAt(0));

	    Object[][] m=new Object[chain_seq.length()][6];
	    
	    for(int i=0; i<chain_seq.length(); i++){
		m[i][0]=new Integer(i+1);
		m[i][1]=new Character(chain_seq.charAt(i));
		m[i][2]=new Integer(i);
		m[i][3]=new Integer(i+2);
		m[i][4]=new Integer(0);
		m[i][5]=new Integer(i+1);
	    }
	    
	    List<Integer> multi_pair=new ArrayList<Integer>();
	    
	    //number of Watson-Crick base pairs
	    int num_WC=0;
	    for(Pair<Integer, String> P : chain_bp){
		if(P.v.equals("WWc")){
		    if((Integer)m[P.left][4]!=0) multi_pair.add(P.left);
		    m[P.left][4]=P.right+1;
		    if((Integer)m[P.right][4]!=0) multi_pair.add(P.right);
		    m[P.right][4]=P.left+1;
		    
		    num_WC+=1;
		}
	    }
	    
	    
	    File ct_fn=new File(SI_DATA_PATH, PDBID+'_'+chainID+".ct");
	    PrintWriter out=new PrintWriter(new BufferedWriter(new FileWriter(ct_fn)));
	    out.printf("%d\t%s_%s\n", chain_seq.length(), PDBID, chainID);
	    for(int i=0; i<chain_seq.length(); i++){
		out.printf("%d\t%s\t%d\t%d\t%d\t%d\n", m[i][0], m[i][1], m[i][2], m[i][3], m[i][4], m[i][5]);
	    }
	    out.close();

	    //if no pairing interaction, RemovePseudoknot does not work correctly.
	    //so we just copy ct to npk.ct, if there is not base pair.
	    if(num_WC==0){
		Files.copy(ct_fn.toPath(), npk_ct_fn.toPath(), StandardCopyOption.REPLACE_EXISTING);
		return;
	    }
	    
	    ProcessBuilder pb = new ProcessBuilder();
	    Map<String, String> env = pb.environment();
	    env.put("DATAPATH", "tools/RNAstructure/data_tables");
	    Process p=Runtime.getRuntime().exec(new File(STAR3D_PATH, "tools/RemovePseudoknots").toString()+" -m "+ct_fn+" "+npk_ct_fn);
	    p.waitFor();
	}
    }
}
